class PromptConstants:
    OUTPUT_DATA_MODEL = "output_data_model"
    TEMPLATE_STR = "template"
    INPUT_VARIABLES_STR = "input_variables"
    PARTIAL_VARIABLES_STR = "partial_variables"
    TEXT_STR = "text"
    OUTPUT_DATA_MODEL_CLASS_SEPARATOR = "\n\n"



import re
import typing

from langchain.output_parsers import PydanticOutputParser
from langchain.schema import OutputParserException

from allms.domain.response import ResponseData, ResponseParsingOutput


class ResponseParser:
    def __init__(self, parser: PydanticOutputParser) -> None:
        self._json_pattern = re.compile(r"{.*?}", re.DOTALL)
        self._parser = parser

    def _clean_extracted_json(self, extracted_json: str) -> str:
        json_without_newlines = extracted_json.replace("\\n", "")
        json_without_backslashes = json_without_newlines.replace("\\", "")

        return json_without_backslashes

    def _extract_json_from_response(self, model_response_data: ResponseData) -> str:
        search_results = self._json_pattern.findall(model_response_data.response)
        
        if len(search_results) == 0:
            return model_response_data.response

        return self._clean_extracted_json(search_results[0])
        
    def _parse_response(
        self, 
        model_response_data: ResponseData
    ) -> ResponseParsingOutput:
        raw_response = self._extract_json_from_response(model_response_data)

        try:
            return ResponseParsingOutput(
                response=self._parser.parse(raw_response), 
                error_message=None
            )
        except OutputParserException as output_parser_exception:
            return ResponseParsingOutput(
                response=None, 
                error_message=f"""
                    An OutputParserException has occurred for the model response: {raw_response}
                    The exception message: {output_parser_exception}
                    """
            )
        
    def parse_model_output(
        self, 
        model_responses_data: typing.List[ResponseData]
    ) -> typing.List[ResponseData]:
        parsed_responses = []

        for model_response_data in model_responses_data:
            if not model_response_data.error:
                response_with_error = self._parse_response(model_response_data)

                parsed_responses.append(ResponseData(
                    input_data=model_response_data.input_data,
                    response=response_with_error.response,
                    error=response_with_error.error_message,
                    number_of_prompt_tokens=model_response_data.number_of_prompt_tokens,
                    number_of_generated_tokens=model_response_data.number_of_generated_tokens

                ))
            else:
                parsed_responses.append(model_response_data)

        return parsed_responses


import logging
import sys


def setup_logger() -> None:
    log_format = '[%(levelname)s] %(asctime)s %(filename)s (%(lineno)d)\t- %(message)s'
    log_dateformat = '%Y-%m-%d %H:%M:%S'
    logging.basicConfig(format=log_format, datefmt=log_dateformat, stream=sys.stdout, level=logging.INFO)



import typing

from pydantic import BaseModel

from allms.domain.input_data import InputData


class ResponseParsingOutput(BaseModel):
    response: typing.Optional[typing.Any]
    error_message: typing.Optional[str]


class ResponseData(BaseModel):
    response: typing.Optional[typing.Any] = None
    input_data: typing.Optional[InputData] = None

    number_of_prompt_tokens: typing.Optional[int] = None
    number_of_generated_tokens: typing.Optional[int] = None
    error: typing.Optional[str] = None

    # Without this, only classes inheriting from the pydantic BaseModel are allowed as field types. Exception isn't
    # such a class and that's why we need it.
    class Config:
        arbitrary_types_allowed = True



class PalmModelDefaults:
    # These values were chosen based on the default values used by the LLM provider
    GCP_MODEL_NAME = "text-bison@001"
    MODEL_TOTAL_MAX_TOKENS = 8192
    MAX_OUTPUT_TOKENS = 1024
    TEMPERATURE = 0.0
    TOP_P = 0.95
    TOP_K = 40
    VERBOSE = True


class GeminiModelDefaults:
    GCP_MODEL_NAME = "gemini-pro"
    MODEL_TOTAL_MAX_TOKENS = 30720
    MAX_OUTPUT_TOKENS = 2048
    TEMPERATURE = 0.0
    TOP_P = 0.95
    TOP_K = 40
    VERBOSE = True


class GemmaModelDefaults:
    GCP_MODEL_NAME = "gemma"
    MODEL_TOTAL_MAX_TOKENS = 8192
    MAX_OUTPUT_TOKENS = 1024
    TEMPERATURE = 0.0
    TOP_P = 0.95
    TOP_K = 40
    VERBOSE = True



from asyncio import AbstractEventLoop
from typing import Optional

from langchain.chat_models import AzureChatOpenAI

from allms.defaults.azure_defaults import AzureGptTurboDefaults
from allms.defaults.general_defaults import GeneralDefaults
from allms.domain.configuration import AzureOpenAIConfiguration
from allms.models.abstract import AbstractModel


class AzureOpenAIModel(AbstractModel):
    def __init__(
            self,
            config: AzureOpenAIConfiguration,
            temperature: float = AzureGptTurboDefaults.TEMPERATURE,
            max_output_tokens: int = AzureGptTurboDefaults.MAX_OUTPUT_TOKENS,
            request_timeout_s: int = AzureGptTurboDefaults.REQUEST_TIMEOUT_S,
            model_total_max_tokens: int = AzureGptTurboDefaults.MODEL_TOTAL_MAX_TOKENS,
            max_concurrency: int = GeneralDefaults.MAX_CONCURRENCY,
            max_retries: int = GeneralDefaults.MAX_RETRIES,
            event_loop: Optional[AbstractEventLoop] = None
    ) -> None:
        self._request_timeout_s = request_timeout_s
        self._config = config

        super().__init__(
            temperature=temperature,
            model_total_max_tokens=model_total_max_tokens,
            max_output_tokens=max_output_tokens,
            max_concurrency=max_concurrency,
            max_retries=max_retries,
            event_loop=event_loop
        )

    def _create_llm(self) -> AzureChatOpenAI:
        return AzureChatOpenAI(
            deployment_name=self._config.deployment,
            api_version=self._config.api_version,
            model_name=self._config.model_name,
            base_url=self._config.base_url,
            api_key=self._config.api_key,
            temperature=self._temperature,
            max_tokens=self._max_output_tokens,
            request_timeout=self._request_timeout_s
        )



import asyncio
import logging
import re
import typing
import urllib
from abc import ABC, abstractmethod
from functools import partial

import google
import openai
from google.api_core.exceptions import InvalidArgument
from langchain.chains import LLMChain
from langchain.chat_models.base import BaseChatModel
from langchain.output_parsers import PydanticOutputParser
from langchain.prompts import ChatPromptTemplate
from langchain_core.language_models.llms import create_base_retry_decorator
from langchain_core.prompts import HumanMessagePromptTemplate, SystemMessagePromptTemplate
from langchain_core.prompts.prompt import PromptTemplate
from pydantic import BaseModel

import allms.exceptions.validation_input_data_exceptions as input_exception_message
import allms.models as models
from allms.chains.long_text_processing_chain import (
    LongTextProcessingChain,
    load_long_text_processing_chain
)
from allms.constants.input_data import IODataConstants
from allms.constants.prompt import PromptConstants
from allms.defaults.general_defaults import GeneralDefaults
from allms.defaults.long_text_chain import LongTextChainDefaults
from allms.domain.enumerables import AggregationLogicForLongInputData, LanguageModelTask
from allms.domain.input_data import InputData
from allms.domain.prompt_dto import SummaryOutputClass, KeywordsOutputClass
from allms.domain.response import ResponseData
from allms.utils.long_text_processing_utils import get_max_allowed_number_of_tokens
from allms.utils.response_parsing_utils import ResponseParser

logger = logging.getLogger(__name__)


class AbstractModel(ABC):
    def __init__(
            self,
            temperature: float,
            max_output_tokens: int,
            model_total_max_tokens: int,
            event_loop: typing.Optional[asyncio.AbstractEventLoop] = None,
            max_concurrency: int = GeneralDefaults.MAX_CONCURRENCY,
            max_retries: int = GeneralDefaults.MAX_RETRIES
    ):
        self._model_total_max_tokens = model_total_max_tokens
        self._max_output_tokens = max_output_tokens
        self._temperature = temperature
        self._semaphore = asyncio.Semaphore(max_concurrency)

        # TODO: To be changed after implementing support for long sequences
        self._task = LanguageModelTask.KEYWORDS
        self._is_long_text_bypass_enabled: bool = False  # Should be false till we fully implement support for long sequences in our package
        self._aggregation_strategy: AggregationLogicForLongInputData = AggregationLogicForLongInputData.SIMPLE_CONCATENATION
        self._parser: typing.Optional[PydanticOutputParser] = None
        self._json_pattern = re.compile(r"{.*?}", re.DOTALL)
        self._is_json_format_injected_into_prompt: bool = True

        if max_output_tokens >= model_total_max_tokens:
            raise ValueError("max_output_tokens has to be lower than model_total_max_tokens")

        self._llm = self._create_llm()

        if not event_loop:
            try:
                event_loop = asyncio.get_running_loop()
            except RuntimeError as error:
                event_loop = asyncio.new_event_loop()
                asyncio.set_event_loop(event_loop)
        self._event_loop = event_loop

        self._predict_example = create_base_retry_decorator(
            error_types=[
                openai.error.RateLimitError, openai.error.APIError, openai.error.Timeout,
                openai.error.APIConnectionError, openai.error.ServiceUnavailableError,
                google.api_core.exceptions.ResourceExhausted, urllib.error.HTTPError
            ],
            max_retries=max_retries,
        )(self._predict_example)

    @abstractmethod
    def _create_llm(self) -> BaseChatModel:
        ...

    def _get_prompt_tokens_number(self, prompt: ChatPromptTemplate, input_data: InputData) -> int:
        return self._llm.get_num_tokens(prompt.format_prompt(**input_data.input_mappings).to_string())

    def _get_model_response_tokens_number(self, model_response: typing.Optional[str]) -> int:
        if model_response:
            return self._llm.get_num_tokens(model_response)
        return 0

    def generate(
            self,
            prompt: str,
            input_data: typing.Optional[typing.List[InputData]] = None,
            output_data_model_class: typing.Optional[typing.Type[BaseModel]] = None,
            system_prompt: typing.Optional[str] = None
    ) -> typing.List[ResponseData]:
        model_responses = self._event_loop.run_until_complete(
            self._generate(
                prompt=prompt,
                input_data=input_data,
                output_data_model_class=output_data_model_class,
                system_prompt=system_prompt
            )
        )

        if output_data_model_class:
            return ResponseParser(self._parser).parse_model_output(model_responses)
        return model_responses

    async def _generate(
            self,
            prompt: str,
            input_data: typing.Optional[typing.List[InputData]] = None,
            output_data_model_class: typing.Optional[typing.Type[BaseModel]] = None,
            system_prompt: typing.Optional[str] = None
    ) -> typing.List[ResponseData]:
        self._validate_system_prompt(system_prompt=system_prompt)
        self._validate_input(prompt=prompt, input_data=input_data)
        if input_data is None:
            # Prompt without symbolic variables is passed - create input_data accordingly
            input_data = [InputData(input_mappings={}, id=IODataConstants.DEFAULT_ID)]

        prompt_template_args = {
            PromptConstants.TEMPLATE_STR: prompt,
            PromptConstants.INPUT_VARIABLES_STR: list(input_data[0].get_input_keys())
        }

        if output_data_model_class:
            self._parser = PydanticOutputParser(pydantic_object=output_data_model_class)

            if self._is_json_format_injected_into_prompt:
                prompt_template_args[PromptConstants.PARTIAL_VARIABLES_STR] = {
                    PromptConstants.OUTPUT_DATA_MODEL: self._parser.get_format_instructions(),
                }
                prompt_template_args[PromptConstants.TEMPLATE_STR] = self._add_output_data_format(prompt=prompt)

        chat_prompts = await self._build_chat_prompts(prompt_template_args, system_prompt)

        prompt_template = ChatPromptTemplate.from_messages(chat_prompts)

        chain = self._get_chain(prompt_template)
        long_chain = self._get_chain_for_long_text(prompt_template)

        predict_example_any_length_partial = partial(
            self._predict_example_of_any_length,
            prompt_template=prompt_template,
            standard_chain=chain,
            long_chain=long_chain
        )

        logger.info("Generating responses...")
        results = list(map(lambda data: predict_example_any_length_partial(input_data=data), input_data))

        responses = await asyncio.gather(*results)

        return responses

    async def _build_chat_prompts(
            self,
            prompt_template_args: dict,
            system_prompt: SystemMessagePromptTemplate
    ) -> typing.List[typing.Union[SystemMessagePromptTemplate, HumanMessagePromptTemplate]]:
        human_message = HumanMessagePromptTemplate(prompt=PromptTemplate(**prompt_template_args))
        if not system_prompt:
            return [human_message]
        system_message_template = SystemMessagePromptTemplate.from_template(system_prompt)

        return [system_message_template, human_message]

    @staticmethod
    def _add_output_data_format(prompt: str) -> str:
        return f"{prompt}{PromptConstants.OUTPUT_DATA_MODEL_CLASS_SEPARATOR}{{{PromptConstants.OUTPUT_DATA_MODEL}}}"

    def _validate_input_data_len(
            self,
            input_data: InputData,
            number_of_prompt_tokens: int
    ):
        if number_of_prompt_tokens > self._model_total_max_tokens:
            raise ValueError(
                f"Prompt is too long. Entire prompt has {number_of_prompt_tokens} tokens, where the max allowed number "
                f"of tokens of the model is {self._model_total_max_tokens}. This leaves no space for the model to "
                f"generate a response and will lead to errors. Example id: {input_data.id}"
            )
        elif number_of_prompt_tokens + self._max_output_tokens > self._model_total_max_tokens:
            logger.warning(
                f"Number of prompt tokens plus generated tokens may exceed the the max allowed number of tokens of the "
                f"model. Entire prompt has {number_of_prompt_tokens} tokens, the max number of tokens to generate is "
                f"{self._max_output_tokens} and the max allowed number of tokens of the model is "
                f"{self._model_total_max_tokens}. Consider lowering the max_output_tokens param or truncating the "
                f"input, because otherwise it may lead to unexpected errors. Example id: {input_data.id}"
            )

    def _predict_example_of_any_length(
            self,
            input_data: InputData,
            prompt_template: ChatPromptTemplate,
            standard_chain: LLMChain,
            long_chain: LLMChain
    ) -> ResponseData:
        number_of_prompt_tokens = self._get_prompt_tokens_number(
            prompt=prompt_template,
            input_data=input_data
        )
        max_token_limit = get_max_allowed_number_of_tokens(self._model_total_max_tokens, self._max_output_tokens)
        is_example_too_long = number_of_prompt_tokens > max_token_limit

        predict_example_partial = partial(
            self._predict_example,
            input_data=input_data,
            prompt_tokens_number=number_of_prompt_tokens
        )
        if is_example_too_long and self._is_long_text_bypass_enabled:
            return predict_example_partial(chain=long_chain)
        return predict_example_partial(chain=standard_chain)

    async def _predict_example(
            self,
            chain: LLMChain,
            input_data: InputData,
            prompt_tokens_number: int
    ) -> ResponseData:
        error_message: typing.Optional[str] = None
        number_of_input_mappings = len(input_data.input_mappings)

        try:
            self._validate_input_data_len(input_data=input_data, number_of_prompt_tokens=prompt_tokens_number)
        except ValueError as value_error:
            logger.info(f"Error for id {input_data.id} has occurred. Message: {value_error} ")
            error_message = f"{IODataConstants.VALUE_ERROR_MESSAGE}: {value_error}"
            return ResponseData(
                input_data=None if number_of_input_mappings == 0 else input_data,
                response=None,
                number_of_prompt_tokens=prompt_tokens_number,
                number_of_generated_tokens=0,
                error=error_message
            )

        try:
            async with self._semaphore:
                if number_of_input_mappings == 0:
                    # Workaround when prompt without symbolic variables is passed - arun() can't be called without any arg
                    model_response = await chain.arun({})
                else:
                    model_response = await chain.arun(**input_data.input_mappings)
        except openai.error.InvalidRequestError as invalid_request_error:
            logger.info(f"Error for id {input_data.id} has occurred. Message: {invalid_request_error} ")
            if invalid_request_error.error.code == "content_filter":
                model_response = None
                error_message = f"{IODataConstants.CONTENT_FILTER_MESSAGE}: {invalid_request_error}"
            else:
                model_response = None
                error_message = f"{IODataConstants.ERROR_MESSAGE_STR}: {invalid_request_error}"

        except (InvalidArgument, ValueError, TimeoutError, openai.error.Timeout) as other_error:
            model_response = None
            logger.info(f"Error for id {input_data.id} has occurred. Message: {other_error} ")
            error_message = f"{type(other_error).__name__}: {other_error}"

        return ResponseData(
            input_data=None if number_of_input_mappings == 0 else input_data,
            response=model_response,
            number_of_prompt_tokens=prompt_tokens_number,
            number_of_generated_tokens=self._get_model_response_tokens_number(model_response),
            error=error_message
        )

    def _get_number_of_tokens_in_prompt(self, prompt: PromptTemplate, input_data: InputData) -> int:
        return self._llm.get_num_tokens(prompt.format_prompt(**input_data.input_mappings).to_string())

    def _get_chain(self, prompt: PromptTemplate) -> LLMChain:
        return LLMChain(
            llm=self._llm,
            prompt=prompt,
        )

    # TODO: When adding support for long documents, we'll need to rethink how output_data_model will be passed to the
    # TODO: aggregation prompt
    def _get_chain_for_long_text(
            self,
            prompt_template: PromptTemplate,
    ) -> LongTextProcessingChain:
        parser = PydanticOutputParser(
            pydantic_object=SummaryOutputClass if self._task == LanguageModelTask.SUMMARY else KeywordsOutputClass)
        reduce_prompt_template = PromptTemplate(
            template=LongTextChainDefaults.AGGREGATION_PROMPT,
            input_variables=["text"],
            partial_variables={PromptConstants.OUTPUT_DATA_MODEL: parser.get_format_instructions()}
        )

        return load_long_text_processing_chain(
            task=self._task,
            llm=self._llm,
            model_total_max_tokens=self._model_total_max_tokens,
            max_output_tokens=self._max_output_tokens,
            map_prompt=prompt_template,
            reduce_prompt=reduce_prompt_template,
            aggregation_strategy=self._aggregation_strategy
        )

    def _validate_input(self, prompt: str, input_data: typing.Optional[typing.List[InputData]] = None) -> None:
        # Extracts text inside the {} but escapes the text inside {{}}
        # This behaviour allows to pass JSON-like strings to the prompt
        # reference: https://github.com/langchain-ai/langchain/issues/1660#issuecomment-1469320129
        prompt_input_variables_set = AbstractModel._extract_input_variables_from_prompt(prompt)
        if PromptConstants.OUTPUT_DATA_MODEL in prompt_input_variables_set:
            prompt_input_variables_set.remove(PromptConstants.OUTPUT_DATA_MODEL)

        if input_data:
            for data in input_data:
                self._validate_input_data(prompt_input_variables_set, data)
        elif len(prompt_input_variables_set) > 0:
            raise ValueError(
                input_exception_message.get_prompt_contains_input_key_when_missing_input_data())

    def _validate_system_prompt(self, system_prompt: typing.Optional[str] = None) -> None:
        if isinstance(self, models.AzureMistralModel) and system_prompt is not None:
            raise ValueError(input_exception_message.get_system_prompt_is_not_supported_by_model())
        elif system_prompt:
            prompt_input_variables_set = AbstractModel._extract_input_variables_from_prompt(system_prompt)
            if prompt_input_variables_set:
                raise ValueError(input_exception_message.get_system_prompt_contains_input_variables())

    @staticmethod
    def _extract_input_variables_from_prompt(prompt: str) -> typing.Set[str]:
        input_variables_pattern = r'(?<!\{)\{([^{}]+)\}(?!\})'
        input_variables_set = set(re.findall(input_variables_pattern, prompt))
        return input_variables_set

    @staticmethod
    def _validate_input_data(
            prompt_input_variables: typing.Set[str],
            input_data: typing.Optional[InputData] = None
    ) -> None:
        if len(input_data.input_mappings.keys()) > 0 and len(prompt_input_variables) == 0:
            raise ValueError(input_exception_message.get_missing_input_data_in_prompt_message(input_data.id))

        if len(input_data.input_mappings.keys()) == 0 and len(prompt_input_variables) > 0:
            raise ValueError(input_exception_message.get_missing_input_data_in_input_data_message(input_data.id))

        if len(input_data.input_mappings.keys()) != len(prompt_input_variables):
            raise ValueError(input_exception_message.get_different_number_of_inputs_message(input_data.id))

        if not prompt_input_variables == set(input_data.get_input_keys()):
            raise ValueError(input_exception_message.get_different_input_keys_message(input_data.id))



import asyncio
import typing
from concurrent.futures import ThreadPoolExecutor

from langchain.callbacks.manager import AsyncCallbackManagerForLLMRun
from langchain_community.chat_models.azureml_endpoint import AzureMLChatOnlineEndpoint


class AzureMLOnlineEndpointAsync(AzureMLChatOnlineEndpoint):

    async def _acall(
            self,
            prompt: str,
            stop: typing.Optional[typing.List[str]] = None,
            run_manager: typing.Optional[AsyncCallbackManagerForLLMRun] = None,
            **kwargs: typing.Any,
    ) -> str:
        # Under the hood, langchain uses urllib.request to query the Azure ML Endpoint. urllib.request is not compatible
        # with asyncio, and that's why we had to implement the function this way
        task_executor = ThreadPoolExecutor()
        return await asyncio.wrap_future(
            task_executor.submit(self._call, prompt, stop, run_manager, **kwargs)
        )



import typing


class IODataConstants:
    TEXT = "text"
    ID = "id"

    PROMPT_TOKENS_NUMBER = "number_of_prompt_tokens"
    GENERATED_TOKENS_NUMBER = "number_of_generated_tokens"

    RESPONSE_STR_NAME = "response"

    ERROR_MESSAGE_STR = "Response error"
    VALUE_ERROR_MESSAGE = "Value Error has occurred"
    INVALID_ARGUMENT_MESSAGE = "Invalid Argument Exception"
    CONTENT_FILTER_MESSAGE = "Content Filter Message"
    TIMEOUT_ERROR_MESSAGE = "Timeout Error"

    SUPPORTED_INPUT_DATA_FORMAT = "csv"
    DEFAULT_ID = "DEFAULT_ID"

    @staticmethod
    def get_columns_for_df_with_responses(input_keys: typing.List[str]) -> typing.List[str]:
        return input_keys + [
            IODataConstants.ID,
            IODataConstants.RESPONSE_STR_NAME,
            IODataConstants.PROMPT_TOKENS_NUMBER,
            IODataConstants.GENERATED_TOKENS_NUMBER
        ]



import typing
from asyncio import AbstractEventLoop

from langchain_community.chat_models.azureml_endpoint import LlamaChatContentFormatter

from allms.defaults.azure_defaults import AzureMistralAIDefaults
from allms.defaults.general_defaults import GeneralDefaults
from allms.domain.configuration import AzureSelfDeployedConfiguration
from allms.models.abstract import AbstractModel
from allms.models.azure_base import AzureMLOnlineEndpointAsync


class AzureMistralModel(AbstractModel):

    def __init__(
            self,
            config: AzureSelfDeployedConfiguration,
            temperature: float = AzureMistralAIDefaults.TEMPERATURE,
            top_p: float = AzureMistralAIDefaults.TOP_P,
            max_output_tokens: int = AzureMistralAIDefaults.MAX_OUTPUT_TOKENS,
            model_total_max_tokens: int = AzureMistralAIDefaults.MODEL_TOTAL_MAX_TOKENS,
            max_concurrency: int = GeneralDefaults.MAX_CONCURRENCY,
            max_retries: int = GeneralDefaults.MAX_RETRIES,
            event_loop: typing.Optional[AbstractEventLoop] = None
    ) -> None:
        self._top_p = top_p
        self._config = config

        super().__init__(
            temperature=temperature,
            model_total_max_tokens=model_total_max_tokens,
            max_output_tokens=max_output_tokens,
            max_concurrency=max_concurrency,
            max_retries=max_retries,
            event_loop=event_loop
        )

        self._is_json_format_injected_into_prompt = False

    def _create_llm(self) -> AzureMLOnlineEndpointAsync:
        model_kwargs = {
            "max_new_tokens": self._max_output_tokens, "top_p": self._top_p, "do_sample": False,
            "return_full_text": False
        }
        if self._temperature > 0:
            model_kwargs["temperature"] = self._temperature
            model_kwargs["do_sample"] = True

        return AzureMLOnlineEndpointAsync(
            endpoint_api_key=self._config.api_key,
            endpoint_url=self._config.endpoint_url,
            model_kwargs=model_kwargs,
            content_formatter=LlamaChatContentFormatter(),
            deployment_name=self._config.deployment
        )



class AzureOpenAIConstants:
    OPENAI_API_TYPE_STR_NAME: str = "OPENAI_API_TYPE"
    OPENAI_API_BASE_STR_NAME: str = "OPENAI_API_BASE"
    OPENAI_API_VERSION_STR_NAME: str = "OPENAI_API_VERSION"
    OPENAI_DEPLOYMENT_NAME_STR_NAME: str = "OPENAI_DEPLOYMENT_NAME"
    OPENAI_API_KEY_STR_NAME: str = "OPENAI_API_KEY"
    OPENAI_MODEL_NAME_STR_NAME: str = "OPENAI_MODEL_NAME"


class AzureMLOnlineEndpointConstants:
    AZURE_API_KEY_STR_NAME = "AZURE_API_KEY"
    AZURE_ENDPOINT_URL_STR_NAME = "AZURE_ENDPOINT_URL"
    AZURE_DEPLOYMENT_NAME_STR_NAME = "AZURE_DEPLOYMENT_NAME"



from pydantic import BaseModel, Field
from typing import List, Union


class SummaryOutputClass(BaseModel):
    summary: str = Field(description="Summary of a product description")


class KeywordsOutputClass(BaseModel):
    keywords: List[str] = Field(description="List of keywords")


class AggregateOutputClass(BaseModel):
    summaries: List[Union[SummaryOutputClass, KeywordsOutputClass]] = Field(description="List of aggregated outputs")



def get_missing_input_data_in_prompt_message(example_id: str) -> str:
    return f"Missing input_keys in the prompt. Error occurred for id={example_id}"


def get_missing_input_data_in_input_data_message(example_id: str) -> str:
    return f"Missing input_keys in the input data. Error occurred for id={example_id}"


def get_different_number_of_inputs_message(example_id: str) -> str:
    return (f"Number of input keys in input_data and prompt are different."
            f"If your intention is to instruct the model to output a JSON, make sure you are using double curly brackets."
            f" Please make sure the input_keys are consistent."
            f" Error has occurred for id={example_id}")


def get_different_input_keys_message(example_id: str) -> str:
    return (f"Input variables in the prompt and in the input_data are different. Please make sure"
            f"the input_keys are consistent. "
            f"If your intention is to instruct the model to output a JSON, make sure you are using double curly brackets."
            f"Error has occurred for id={example_id}")


def get_prompt_contains_input_key_when_missing_input_data() -> str:
    return f"When no input_data is provided prompt cannot contain any input_key."


def get_system_prompt_contains_input_variables() -> str:
    return "System prompt cannot contain any input variables. Please fix your system message and try again."


def get_system_prompt_is_not_supported_by_model() -> str:
    return "Mistral-based models don't support `system_prompt` parameter."



from typing import List, Optional, Any, Dict

from google.cloud.aiplatform.models import Prediction
from langchain_community.llms.vertexai import VertexAI, VertexAIModelGarden
from langchain_core.callbacks import AsyncCallbackManagerForLLMRun
from langchain_core.outputs import LLMResult, Generation
from pydash import chain

from allms.constants.vertex_ai import VertexModelConstants


class CustomVertexAI(VertexAI):
    async def _agenerate(
        self,
        prompts: List[str],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> LLMResult:
        def was_response_blocked(generation: Generation) -> bool:
            return (
                generation.generation_info is not None 
                    and 'is_blocked' in generation.generation_info 
                    and generation.generation_info['is_blocked']
            )

        result = await super()._agenerate(
            prompts=prompts,
            stop=stop,
            run_manager=run_manager,
            **kwargs
        )

        return LLMResult(
            generations=(
                chain(result.generations)
                .map(lambda generation_candidates: (
                    chain(generation_candidates)
                    .map(
                        lambda single_candidate: Generation(
                            text=VertexModelConstants.RESPONSE_BLOCKED_STR
                        ) if was_response_blocked(single_candidate) else single_candidate
                    )
                    .value()
                ))
                .value()
            ),
            llm_output=result.llm_output,
            run=result.run
        )


class VertexAIModelGardenWrapper(VertexAIModelGarden):
    temperature: float = 0.0
    max_tokens: int = 128
    top_p: float = 0.95
    top_k: int = 40
    n: int = 1

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.allowed_model_args = list(self._default_params.keys())

    @property
    def _default_params(self) -> Dict[str, Any]:
        return {
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            "top_k": self.top_k,
            "top_p": self.top_p,
            "n": self.n
        }

    def _parse_response(self, predictions: "Prediction", prompts: List[str]) -> LLMResult:
        generations: List[List[Generation]] = []
        for result, prompt in zip(predictions.predictions, prompts):
            if isinstance(result, str):
                generations.append([Generation(text=self._parse_prediction(result, prompt))])
            else:
                generations.append(
                    [
                        Generation(text=self._parse_prediction(prediction, prompt))
                        for prediction in result
                    ]
                )
        return LLMResult(generations=generations)

    def _parse_prediction(self, prediction: Any, prompt: str) -> str:
        parsed_prediction = super()._parse_prediction(prediction)
        try:
            text_to_remove = f"Prompt:\n{prompt}\nOutput:\n"
            return parsed_prediction.rsplit(text_to_remove, maxsplit=1)[1]
        except Exception:
            raise ValueError(f"Output returned from the model doesn't follow the expected format.")

    async def _agenerate(
        self,
        prompts: List[str],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> LLMResult:
        kwargs = {**kwargs, **self._default_params}
        instances = self._prepare_request(prompts, **kwargs)
        response = await self.async_client.predict(
            endpoint=self.endpoint_path, instances=instances
        )
        return self._parse_response(response, prompts)




class VertexModelConstants:
    RESPONSE_SEPARATOR = "<LLM_RESPONSE_SEPARATOR>"
    RESPONSE_BLOCKED_STR = "<RESPONSE_BLOCKED>"

    GCP_PROJECT_ID_STR_NAME = "GCP_PROJECT_ID"
    GCP_LLM_REGION_STR_NAME = "GCP_LLM_REGION"
    GCP_MODEL_NAME_STR_NAME = "GCP_MODEL_NAME"



from asyncio import AbstractEventLoop

from langchain_community.llms.vertexai import VertexAIModelGarden
from typing import Optional

from allms.defaults.general_defaults import GeneralDefaults
from allms.defaults.vertex_ai import GemmaModelDefaults
from allms.domain.configuration import VertexAIModelGardenConfiguration
from allms.models.vertexai_base import VertexAIModelGardenWrapper
from allms.models.abstract import AbstractModel


class VertexAIGemmaModel(AbstractModel):
    def __init__(
            self,
            config: VertexAIModelGardenConfiguration,
            temperature: float = GemmaModelDefaults.TEMPERATURE,
            top_k: int = GemmaModelDefaults.TOP_K,
            top_p: float = GemmaModelDefaults.TOP_P,
            max_output_tokens: int = GemmaModelDefaults.MAX_OUTPUT_TOKENS,
            model_total_max_tokens: int = GemmaModelDefaults.MODEL_TOTAL_MAX_TOKENS,
            max_concurrency: int = GeneralDefaults.MAX_CONCURRENCY,
            max_retries: int = GeneralDefaults.MAX_RETRIES,
            verbose: bool = GemmaModelDefaults.VERBOSE,
            event_loop: Optional[AbstractEventLoop] = None
    ) -> None:
        self._top_p = top_p
        self._top_k = top_k
        self._verbose = verbose
        self._config = config

        super().__init__(
            temperature=temperature,
            model_total_max_tokens=model_total_max_tokens,
            max_output_tokens=max_output_tokens,
            max_concurrency=max_concurrency,
            max_retries=max_retries,
            event_loop=event_loop
        )

        self._is_json_format_injected_into_prompt = False

    def _create_llm(self) -> VertexAIModelGarden:
        return VertexAIModelGardenWrapper(
            model_name=GemmaModelDefaults.GCP_MODEL_NAME,
            max_tokens=self._max_output_tokens,
            temperature=self._temperature,
            top_p=self._top_p,
            top_k=self._top_k,
            verbose=self._verbose,
            project=self._config.cloud_project,
            location=self._config.cloud_location,
            endpoint_id=self._config.endpoint_id
        )



from enum import Enum
from typing import List


class ListConvertableEnum(Enum):
    @classmethod
    def get_values(cls) -> List[str]:
        return list(map(lambda field: field.value, cls))


class AggregationLogicForLongInputData(str, ListConvertableEnum):
    SIMPLE_CONCATENATION = "SIMPLE_CONCATENATION"
    REDUCE_BY_LLM_PROMPTING = "REDUCE_BY_LLM_PROMPTING"


class AvailableModels(str, ListConvertableEnum):
    AZURE_OPENAI_MODEL = "azure_openai"
    AZURE_LLAMA2_MODEL = "azure_llama2"
    AZURE_MISTRAL_MODEL = "azure_mistral"
    VERTEXAI_PALM2_MODEL = "vertexai_palm2"
    VERTEXAI_GEMINI_MODEL = "vertexai_gemini"
    VERTEXAI_GEMMA_MODEL = "vertexai_gemma"


class LanguageModelTask(str, ListConvertableEnum):
    SUMMARY = "SUMMARY"
    KEYWORDS = "KEYWORDS"



from typing import Dict, Type

from allms.domain.enumerables import AvailableModels
from allms.models.abstract import AbstractModel
from allms.models.azure_llama2 import AzureLlama2Model
from allms.models.azure_mistral import AzureMistralModel
from allms.models.azure_openai import AzureOpenAIModel
from allms.models.vertexai_gemini import VertexAIGeminiModel
from allms.models.vertexai_palm import VertexAIPalmModel
from allms.models.vertexai_gemma import VertexAIGemmaModel

__all__ = [
    "AzureOpenAIModel",
    "AzureLlama2Model",
    "AzureMistralModel",
    "VertexAIPalmModel",
    "VertexAIGeminiModel",
    "VertexAIGemmaModel",
    "get_available_models"
]


def get_available_models() -> Dict[str, Type[AbstractModel]]:
    return {
        AvailableModels.AZURE_OPENAI_MODEL: AzureOpenAIModel,
        AvailableModels.AZURE_LLAMA2_MODEL: AzureLlama2Model,
        AvailableModels.AZURE_MISTRAL_MODEL: AzureMistralModel,
        AvailableModels.VERTEXAI_PALM2_MODEL: VertexAIPalmModel,
        AvailableModels.VERTEXAI_GEMINI_MODEL: VertexAIGeminiModel,
        AvailableModels.VERTEXAI_GEMMA_MODEL: VertexAIGemmaModel,
    }




class LongTextChainDefaults:
    OVERLAP_SIZE = 50
    AGGREGATOR_DEFAULT_STR_SEPARATOR = ", "

    # TODO Refactor along with adding support for LongDocument processing
    AGGREGATION_PROMPT = """You're an AI agent that combines product summaries. Write a summary of the provided summaries. Keep the most important information and discard
            redundant information. The input is a JSON object.

            {output_data_model}

            {text}
        """



import csv
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, OrderedDict

import fsspec

from allms.constants.input_data import IODataConstants
from allms.domain.input_data import InputData

logger = logging.getLogger(__name__)


def load_csv(
        path: str,
        limit: Optional[int] = None
) -> List[OrderedDict[Any, Any]]:
    logger.info(f"Loading test data from {path}")
    with open(path, mode='r') as csv_file:
        csv_reader = csv.DictReader(csv_file)
        data = list(csv_reader)
        return data[:limit] if limit else data


def load_csv_to_input_data(path: str, limit: Optional[int] = None) -> List[InputData]:
    csv_data = load_csv(path, limit=limit)
    return list(
        map(
            lambda row: InputData(input_mappings=drop_dict_key(row, IODataConstants.ID),
                                  id=str(row[IODataConstants.ID])),
            csv_data
        )
    )


def drop_dict_key(dictionary: Dict[Any, Any], key: Any) -> Dict[Any, Any]:
    dict_copy = dictionary.copy()
    dict_copy.pop(key)
    return dict_copy


def load_credentials(path: Union[str, Path]) -> str:
    with fsspec.open(path, "r") as credentials_file:
        return credentials_file.readline()



import typing

from pydantic import BaseModel


class InputData(BaseModel):
    input_mappings: typing.Dict[str, str]
    id: str

    def get_input_keys(self) -> typing.List[str]:
        return list(self.input_mappings.keys())



import asyncio
from functools import reduce
from typing import List, Any, Tuple, Optional, Union

from langchain import LLMChain, BasePromptTemplate
from langchain.base_language import BaseLanguageModel
from langchain.chains.combine_documents.base import BaseCombineDocumentsChain
from langchain.schema import Document

from allms.domain.enumerables import AggregationLogicForLongInputData, LanguageModelTask
from allms.domain.input_data import InputData
from allms.domain.prompt_dto import (AggregateOutputClass, KeywordsOutputClass, SummaryOutputClass)
from allms.utils.long_text_processing_utils import split_text_to_max_size


class LongTextProcessingChain(BaseCombineDocumentsChain):
    task: LanguageModelTask
    model_total_max_tokens: int
    max_output_tokens: int
    map_llm_chain: LLMChain
    reduce_llm_chain: LLMChain
    input_data_variable_name: str
    aggregation_strategy: AggregationLogicForLongInputData

    @property
    def _chain_type(self) -> str:
        return "long_description_chain"

    async def combine_docs(self, input_data: Document, **kwargs: Any) -> Tuple[str, dict]:
        chunked_input: List[Document] = split_text_to_max_size(
            llm=self.map_llm_chain.llm,
            prompt_template=self.map_llm_chain.prompt,
            text=input_data,
            model_total_max_tokens=self.model_total_max_tokens,
            max_output_tokens=self.max_output_tokens
        )

        chunk_responses = await self._map_step(chunked_input)
        aggregated_response = self._reduce_step(chunk_responses)

        return aggregated_response, {}

    async def acombine_docs(self, input_data: List[Document], **kwargs: Any) -> Tuple[str, dict]:
        return await self.combine_docs(input_data)

    async def _map_step(self, chunked_document: List[Document]) -> List[str]:
        results = list(map(lambda document: self.map_llm_chain.arun(document), chunked_document))
        return await asyncio.gather(*results)

    def _reduce_step(self, chunk_responses: List[InputData]) -> str:
        if self.aggregation_strategy == AggregationLogicForLongInputData.REDUCE_BY_LLM_PROMPTING:
            return self._construct_input_from_list_and_run_reduce_chain(chunk_responses)
        elif self.aggregation_strategy == AggregationLogicForLongInputData.SIMPLE_CONCATENATION:
            if self.task == LanguageModelTask.SUMMARY:
                return self._aggregate_results_for_summary(chunk_responses).json()
            elif self.task == LanguageModelTask.KEYWORDS:
                return self._aggregate_results_for_keywords(chunk_responses).json()

    def _deserialize_response(self, response: str) -> Union[SummaryOutputClass, KeywordsOutputClass]:
        if self.task == LanguageModelTask.SUMMARY:
            return SummaryOutputClass.parse_raw(response)
        elif self.task == LanguageModelTask.KEYWORDS:
            return KeywordsOutputClass.parse_raw(response)

    def _construct_input_from_list_and_run_reduce_chain(self, response_list: List[InputData]) -> str:
        aggregate_input = Document(
            page_content=AggregateOutputClass(summaries=[
                self._deserialize_response(response) for response in response_list]
            ).json()
        )

        return self.reduce_llm_chain.run(aggregate_input.text)

    @staticmethod
    def _aggregate_results_for_summary(chunk_responses: List[Document]) -> SummaryOutputClass:
        return SummaryOutputClass(summary=" ".join([
            SummaryOutputClass.parse_raw(response_json).summary for response_json in chunk_responses
        ]))

    @staticmethod
    def _aggregate_results_for_keywords(chunk_responses: List[str]) -> KeywordsOutputClass:
        return KeywordsOutputClass(keywords=list(reduce(
            lambda x, y: x + y,
            [KeywordsOutputClass.parse_raw(response_json).keywords for response_json in chunk_responses],
            []
        )))


def load_long_text_processing_chain(
        task: LanguageModelTask,
        llm: BaseLanguageModel,
        model_total_max_tokens: int,
        max_output_tokens: int,
        map_prompt: BasePromptTemplate,
        reduce_prompt: BasePromptTemplate,
        aggregation_strategy: AggregationLogicForLongInputData,
        input_data_variable_name: str = "text",
        verbose: Optional[bool] = None
) -> LongTextProcessingChain:
    map_chain = LLMChain(llm=llm, prompt=map_prompt, verbose=verbose)
    reduce_chain = LLMChain(llm=llm, prompt=reduce_prompt, verbose=verbose)

    return LongTextProcessingChain(
        task=task,
        model_total_max_tokens=model_total_max_tokens,
        max_output_tokens=max_output_tokens,
        map_llm_chain=map_chain,
        reduce_llm_chain=reduce_chain,
        input_data_variable_name=input_data_variable_name,
        aggregation_strategy=aggregation_strategy,
        verbose=verbose
    )



from dataclasses import dataclass
from typing import Optional

from allms.defaults.vertex_ai import GeminiModelDefaults, PalmModelDefaults


@dataclass
class AzureOpenAIConfiguration:
    base_url: str
    deployment: str
    model_name: str
    api_version: str
    api_key: str


@dataclass
class AzureSelfDeployedConfiguration:
    api_key: str
    deployment: str
    endpoint_url: str


@dataclass
class VertexAIConfiguration:
    cloud_project: str
    cloud_location: str
    palm_model_name: Optional[str] = PalmModelDefaults.GCP_MODEL_NAME
    gemini_model_name: Optional[str] = GeminiModelDefaults.GCP_MODEL_NAME


class VertexAIModelGardenConfiguration(VertexAIConfiguration):
    def __init__(
        self,
        cloud_project: str,
        cloud_location: str,
        endpoint_id: str
    ):
        super().__init__(
            cloud_project=cloud_project,
            cloud_location=cloud_location,
            palm_model_name=None,
            gemini_model_name=None
        )
        self.endpoint_id = endpoint_id



from dataclasses import dataclass


@dataclass
class AzureGptTurboDefaults:
    OPENAI_API_TYPE: str = "azure"
    OPENAI_MODEL_NAME: str = "gpt-3.5-turbo"

    # These values were chosen based on the default values used by the LLM provider
    MODEL_TOTAL_MAX_TOKENS: int = 4096
    TEMPERATURE = 0.0
    MAX_OUTPUT_TOKENS: int = 512
    REQUEST_TIMEOUT_S = 60


@dataclass
class AzureLlama2Defaults:
    MODEL_TOTAL_MAX_TOKENS: int = 4096
    MAX_OUTPUT_TOKENS: int = 512
    TEMPERATURE = 0.0
    TOP_P = 1.0


@dataclass
class AzureMistralAIDefaults:
    MODEL_TOTAL_MAX_TOKENS: int = 8192
    MAX_OUTPUT_TOKENS: int = 1024
    TEMPERATURE = 0.0
    TOP_P = 1.0



from typing import List

from langchain import BasePromptTemplate
from langchain.base_language import BaseLanguageModel
from langchain.schema import Document

from allms.defaults.long_text_chain import LongTextChainDefaults


def truncate_text_to_max_size(
        llm: BaseLanguageModel,
        prompt_template: BasePromptTemplate,
        text: str,
        model_total_max_tokens: int,
        max_output_tokens: int,
) -> str:
    """
    This function is supposed to truncate the input to fit the maximum context size of a model. The problem is that
    the max context size is in tokens and in our code we operate on raw, un-tokenized strings. We can only calculate
    how many tokens given string has. So to find the point on which we should truncate, this function calculates in
    tokens how many times the current `text` is longer than the allowed limit. Then it assumes that the ration is true
    also when reasoning on words instead of tokens. And based on this the `split_point_index` is calculated.
    This is only an approximation (ration calculated on tokens is only similar to the ratio calculated on words, but in
    most of the cases it won't be the same). That's why this function is used in a recursive way. It calculates the
    split point, truncates the text and checks again if the total prompt length is lower than the max context size of
    a model. If not it reruns itself again and if yes, then it returns the truncated text.

    Another possibility would be to implement this function using tokenizer to tokenize text to tokens, then truncate
    the text, detokenize it to string and return truncated text. But for this solution, first we'd need to have a list
    of tokenizers used by every model we'd like to support (now it's provided inside langchain) and second, the
    tokenization and de-tokenization steps could change the input prompt by introducing some artifacts.
    """
    max_token_limit = get_max_allowed_number_of_tokens(model_total_max_tokens, max_output_tokens)
    num_tokens = int(llm.get_num_tokens(prompt_template.format(text=text)))

    if num_tokens <= max_token_limit:
        return text

    # We add `text="text"` and not empty string, because the empty string may be tokenized together with the whitespaces
    # that are around it in the prompt. But when joining the actual `{text}` with the prompt instructions we get one
    # additional token
    num_tokens_prompt_wo_text = int(llm.get_num_tokens(prompt_template.format(text="text")))
    num_tokens_text = int(llm.get_num_tokens(text))
    num_tokens_left_for_text = max_token_limit - num_tokens_prompt_wo_text
    if num_tokens_left_for_text <= 0:
        raise ValueError("Prompt instruction (without the actual text) is longer than the allowed model input length")

    # How many times the current text is longer than the allowed length
    current_to_allowed_length_ration = num_tokens_text / num_tokens_left_for_text
    words = text.split()
    split_point_index = int(len(words) / current_to_allowed_length_ration)

    text_truncated = " ".join(words[:split_point_index])

    return truncate_text_to_max_size(
        llm=llm,
        prompt_template=prompt_template,
        text=text_truncated,
        model_total_max_tokens=model_total_max_tokens,
        max_output_tokens=max_output_tokens
    )


def split_text_to_max_size(
        llm: BaseLanguageModel,
        prompt_template: BasePromptTemplate,
        text: str,
        model_total_max_tokens: int,
        max_output_tokens: int,
        overlap_size: int = LongTextChainDefaults.OVERLAP_SIZE
) -> List[Document]:
    max_token_limit = get_max_allowed_number_of_tokens(model_total_max_tokens, max_output_tokens)
    if int(llm.get_num_tokens(prompt_template.format(text=text))) < max_token_limit:
        return [Document(page_content=text)]

    words = text.split()
    middle_word_index = len(words) // 2

    overlap_left = overlap_size // 2
    overlap_right = overlap_size - overlap_left
    data_left_half = " ".join(words[:middle_word_index + overlap_left])
    data_right_half = " ".join(words[middle_word_index - overlap_right:])

    return (
            split_text_to_max_size(llm=llm, prompt_template=prompt_template, text=data_left_half,
                                   model_total_max_tokens=model_total_max_tokens, max_output_tokens=max_output_tokens)
            + split_text_to_max_size(llm=llm, prompt_template=prompt_template, text=data_right_half,
                                     model_total_max_tokens=model_total_max_tokens, max_output_tokens=max_output_tokens)
    )


def get_max_allowed_number_of_tokens(model_total_max_tokens: int, max_output_tokens: int) -> int:
    buffer = 50  # for things like BOS, EOS and other unexpected things
    return model_total_max_tokens - max_output_tokens - buffer



from asyncio import AbstractEventLoop
from langchain_community.llms.vertexai import VertexAI
from typing import Optional

from allms.defaults.general_defaults import GeneralDefaults
from allms.defaults.vertex_ai import PalmModelDefaults
from allms.domain.configuration import VertexAIConfiguration
from allms.models.vertexai_base import CustomVertexAI
from allms.models.abstract import AbstractModel


class VertexAIPalmModel(AbstractModel):
    def __init__(
            self,
            config: VertexAIConfiguration,
            temperature: float = PalmModelDefaults.TEMPERATURE,
            top_k: int = PalmModelDefaults.TOP_K,
            top_p: float = PalmModelDefaults.TOP_P,
            max_output_tokens: int = PalmModelDefaults.MAX_OUTPUT_TOKENS,
            model_total_max_tokens: int = PalmModelDefaults.MODEL_TOTAL_MAX_TOKENS,
            max_concurrency: int = GeneralDefaults.MAX_CONCURRENCY,
            max_retries: int = GeneralDefaults.MAX_RETRIES,
            verbose: bool = PalmModelDefaults.VERBOSE,
            event_loop: Optional[AbstractEventLoop] = None
    ) -> None:
        self._top_p = top_p
        self._top_k = top_k
        self._verbose = verbose
        self._config = config

        super().__init__(
            temperature=temperature,
            model_total_max_tokens=model_total_max_tokens,
            max_output_tokens=max_output_tokens,
            max_concurrency=max_concurrency,
            max_retries=max_retries,
            event_loop=event_loop
        )

    def _create_llm(self) -> VertexAI:
        return CustomVertexAI(
            model_name=self._config.palm_model_name,
            max_output_tokens=self._max_output_tokens,
            temperature=self._temperature,
            top_p=self._top_p,
            top_k=self._top_k,
            verbose=self._verbose,
            project=self._config.cloud_project,
            location=self._config.cloud_location
        )